"""
Note: this test file was heavily influenced by its dbplyr counterpart.

https://github.com/tidyverse/dbplyr/blob/master/tests/testthat/test-verb-mutate.R
"""
    
from siuba import _, mutate, select, group_by, summarize, filter
from siuba.dply.vector import row_number

import pytest
from .helpers import assert_equal_query, data_frame, backend_notimpl, backend_sql
from string import ascii_lowercase 

DATA = data_frame(a = [1,2,3], b = [9,8,7])

@pytest.fixture(scope = "module")
def dfs(backend):
    return backend.load_df(DATA)

@pytest.mark.parametrize("query, output", [
    (mutate(x = _.a + _.b), DATA.assign(x = [10, 10, 10])),
    pytest.param( mutate(x = _.a + _.b) >> summarize(ttl = _.x.sum()), data_frame(ttl = 30.0), marks = pytest.mark.skip("TODO: failing sqlite?")),
    (mutate(x = _.a + 1, y = _.b - 1), DATA.assign(x = [2,3,4], y = [8,7,6])),
    (mutate(x = _.a + 1) >> mutate(y = _.b - 1), DATA.assign(x = [2,3,4], y = [8,7,6])),
    (mutate(x = _.a + 1, y = _.x + 1), DATA.assign(x = [2,3,4], y = [3,4,5]))
    ])
def test_mutate_basic(dfs, query, output):
    assert_equal_query(dfs, query, output)

@pytest.mark.parametrize("query, output", [
    (mutate(x = 1), DATA.assign(x = 1)),
    (mutate(x = "a"), DATA.assign(x = "a")),
    (mutate(x = 1.2), DATA.assign(x = 1.2))
    ])
def test_mutate_literal(dfs, query, output):
    assert_equal_query(dfs, query, output)


def test_select_mutate_filter(dfs):
    assert_equal_query(
            dfs,
            select(_.x == _.a) >> mutate(y = _.x * 2) >> filter(_.y == 2),
            data_frame(x = 1, y = 2)
            )

@pytest.mark.skip("TODO: check most recent vars for efficient mutate (#41)")
def test_mutate_smart_nesting(dfs):
    # y and z both use x, so should create only 1 extra query
    lazy_tbl = dfs >> mutate(x = _.a + 1, y = _.x + 1, z = _.x + 1)

    query = lazy_tbl.last_op.fromclause

    assert query is lazy_tbl.ops[0]
    assert isinstance(query.fromclause, sqlalchemy.Table )


@pytest.mark.skip("TODO: does pandas backend preserve order? (#42)")
def test_mutate_reassign_column_ordering(dfs):
    assert_equal_query(
            dfs,
            mutate(c = 3, a = 1, b = 2),
            data_frame(a = 1, b = 2, c = 3)
            )


@backend_sql
@backend_notimpl("sqlite")
def test_mutate_window_funcs(backend):
    data = data_frame(x = range(1, 5), g = [1,1,2,2])
    dfs = backend.load_df(data)
    assert_equal_query(
            dfs,
            group_by(_.g) >> mutate(row_num = row_number(_).astype(float)),
            data.assign(row_num = [1.0, 2, 1, 2])
            )


@backend_notimpl("sqlite")
def test_mutate_using_agg_expr(backend):
    data = data_frame(x = range(1, 5), g = [1,1,2,2])
    dfs = backend.load_df(data)
    assert_equal_query(
            dfs,
            group_by(_.g) >> mutate(y = _.x - _.x.mean()),
            data.assign(y = [-.5, .5, -.5, .5])
            )

@backend_sql # TODO: pandas outputs a int column
@backend_notimpl("sqlite")
def test_mutate_using_cuml_agg(backend):
    data = data_frame(x = range(1, 5), g = [1,1,2,2])
    dfs = backend.load_df(data)

    # cuml window without arrange before generates warning
    with pytest.warns(None):
        assert_equal_query(
                dfs,
                group_by(_.g) >> mutate(y = _.x.cumsum()),
                data.assign(y = [1.0, 3, 3, 7])
                )

def test_mutate_overwrites_prev(backend):
    # TODO: check that query doesn't generate a CTE
    dfs = backend.load_df(data_frame(x = range(1, 5), g = [1,1,2,2]))
    assert_equal_query(
            dfs,
            mutate(x = _.x + 1) >> mutate(x = _.x + 1),
            data_frame(x = [3,4,5,6], g = [1,1,2,2])
            )






#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copied from https://github.com/spatialaudio/nbsphinx/blob/0.2.7/doc/conf.py

# Select nbsphinx and, if needed, add a math extension (mathjax or pngmath):
extensions = [
    'nbsphinx',
    #'sphinx.ext.mathjax',
]

# Exclude build directory and Jupyter backup files:
exclude_patterns = ['_build', '**.ipynb_checkpoints', '**.swp']

# Default language for syntax highlighting in reST and Markdown cells
highlight_language = 'none'

# -- These set defaults that can be overridden through notebook metadata --

# See http://nbsphinx.readthedocs.org/en/latest/allow-errors.html
# and http://nbsphinx.readthedocs.org/en/latest/timeout.html for more details.

# If True, the build process is continued even if an exception occurs:
#nbsphinx_allow_errors = True

# Controls when a cell will time out (defaults to 30; use -1 for no timeout):
#nbsphinx_timeout = 60

# Default Pygments lexer for syntax highlighting in code cells
#nbsphinx_codecell_lexer = 'ipython3'

# -- The settings below this line are not specific to nbsphinx ------------

master_doc = 'index'

project = 'siuba'
author = 'Michael Chow'
copyright = '2019, ' + author

linkcheck_ignore = [r'http://localhost:\d+/']

# -- Get version information from Git -------------------------------------

# TODO: use mock to pull version info
try:
    from subprocess import check_output
    release = check_output(['git', 'describe', '--tags', '--always'])
    release = release.decode().strip()
except Exception:
    release = '<unknown>'

# -- Options for HTML output ----------------------------------------------

import alabaster

html_title = project + ' version ' + release

html_theme = 'alabaster'

html_theme_options = {
    #"description": "",
    "github_user": "machow",
    "github_repo": "siuba",
    "fixed_sidebar": False,
    "github_banner": True,
    "github_button": False
    }

# -- nbsphinx customization ---------------------------------------------------

import jupytext


nbsphinx_custom_formats = {
    '.Rmd': lambda s: jupytext.reads(s, '.Rmd'),
}

# hide prompt numbers. we change pd display options in a hidden cell, so it looks
# funny to start at [2]
nbsphinx_prompt_width = 0

# This is processed by Jinja2 and inserted before each notebook
nbsphinx_epilog = r"""
{% set docname = env.doc2path(env.docname, base='docs') %}


.. only:: html

    .. role:: raw-html(raw)
        :format: html

    .. nbinfo::

        Edit page on github `here`__.
        Interactive version:
        :raw-html:`<a href="https://mybinder.org/v2/gh/machow/siuba/{{ env.config.release }}?filepath={{ docname }}"><img alt="Binder badge" src="https://mybinder.org/badge_logo.svg" style="vertical-align:text-bottom"></a>`

    __ https://github.com/machow/siuba/blob/
        {{ env.config.release }}/{{ docname }}

"""


# -- Options for LaTeX output ---------------------------------------------

#latex_elements = {
#    'papersize': 'a4paper',
#    'preamble': r"""
#\usepackage{lmodern}  % heavier typewriter font
#""",
#}
#
#latex_documents = [
#    (master_doc, 'nbsphinx.tex', project, author, 'howto'),
#]
#
#latex_show_urls = 'footnote'



import pandas as pd
import numpy as np
from functools import singledispatch
from ..siu import Symbolic, create_sym_call,Call


def register_symbolic(f):
    # TODO: don't use singledispatch if it has already been done
    f = singledispatch(f)
    @f.register(Symbolic)
    def _dispatch_symbol(__data, *args, **kwargs):
        return create_sym_call(f, __data.source, *args, **kwargs)

    return f

def _expand_bool(x, f):
    return x.expanding().apply(f).astype(bool)

@register_symbolic
def cumall(x):
    return _expand_bool(x, np.all)


@register_symbolic
def cumany(x):
    return _expand_bool(x, np.any)


@register_symbolic
def cummean(x):
    return x.expanding().mean()

@register_symbolic
def desc(x):
    return x.sort_values()


@register_symbolic
def dense_rank(x):
    return x.rank(method = "dense")


@register_symbolic
def percent_rank(x):
    NotImplementedError("PRs welcome")


@register_symbolic
def min_rank(x):
    return x.rank(method = "min")


@register_symbolic
def cume_dist(x):
    return x.rank(method = "max") / x.count()


@register_symbolic
def row_number(x):
    if isinstance(x, pd.DataFrame):
        n = x.shape[0]
    else:
        n = len(x)
    return np.arange(1, n + 1)


@register_symbolic
def ntile(x, n):
    NotImplementedError("ntile not implemented")


@register_symbolic
def between(x, left, right):
    # note: NA -> False, in tidyverse NA -> NA
    return x.between(left, right)
    

@register_symbolic
def coalesce(*args):
    NotImplementedError("coalesce not implemented")


@register_symbolic
def lead(x, n = 1, default = None):
    res = x.shift(-1*n)

    if default is not None:
        res.iloc[-n:] = default

    return res


@register_symbolic
def lag(x, n = 1, default = None):
    res = x.shift(n)

    if default is not None:
        res.iloc[:n] = default

    return res


@register_symbolic
def n(x):
    if isinstance(x, pd.DataFrame):
        return x.shape[0]

    return len(x)


@register_symbolic
def n_distinct(x):
    return len(x.unique())


@register_symbolic
def na_if(x, y):
    y = [y] if not np.ndim(y) else y

    tmp_x = x.copy(deep = True)
    tmp_x[x.isin(y)] = np.nan

    return tmp_x


@register_symbolic
def near():
    NotImplementedError("near not implemented") 


@register_symbolic
def nth():
    NotImplementedError("nth not implemented") 


@register_symbolic
def first():
    NotImplementedError("first not implemented")


@register_symbolic
def last():
    NotImplementedError("last not implemented")



"""
Note: this test file was heavily influenced by its dbplyr counterpart.

https://github.com/tidyverse/dbplyr/blob/master/tests/testthat/test-verb-mutate.R
"""
    
from siuba import _, mutate, select, group_by, summarize, filter
from siuba.dply.vector import row_number, n

import pytest
from .helpers import assert_equal_query, data_frame, backend_notimpl, backend_sql
from string import ascii_lowercase 

DATA = data_frame(x = [1,2,3,4], g = ['a', 'a', 'b', 'b'])

@pytest.fixture(scope = "module")
def df(backend):
    return backend.load_df(DATA)

@pytest.fixture(scope = "module")
def df_float(backend):
    return backend.load_df(DATA.assign(x = lambda d: d.x.astype(float)))

@pytest.fixture(scope = "module")
def gdf(df):
    return df >> group_by(_.g)


@pytest.mark.parametrize("query, output", [
    (summarize(y = n(_)), data_frame(y = 4)),
    (summarize(y = _.x.min()), data_frame(y = 1)),
    ])
def test_summarize_ungrouped(df, query, output):
    assert_equal_query(df, query, output)


@pytest.mark.skip("TODO: should return 1 row (#63)")
def test_ungrouped_summarize_literal(df, query, output):
    assert_equal_query(df, summarize(y = 1), data_frame(y = 1)) 


@backend_notimpl("sqlite")
def test_summarize_after_mutate_cuml_win(backend, df_float):
    assert_equal_query(
            df_float,
            mutate(y = _.x.cumsum()) >> summarize(z = _.y.max()),
            data_frame(z = [10.])
            )


@backend_sql
def test_summarize_keeps_group_vars(backend, gdf):
    q = gdf >> summarize(n = n(_))
    assert list(q.last_op.c.keys()) == ["g", "n"]


@pytest.mark.parametrize("query, output", [
    (summarize(y = 1), data_frame(g = ['a', 'b'], y = [1, 1])),
    (summarize(y = n(_)), data_frame(g = ['a', 'b'], y = [2,2])),
    (summarize(y = _.x.min()), data_frame(g = ['a', 'b'], y = [1, 3])),
    # TODO: same issue as above
    #(mutate(y = _.x.cumsum()) >> summarize(z = _.y.max()), data_frame(y = [3, 7]))
    ])
def test_summarize_grouped(gdf, query, output):
    assert_equal_query(gdf, query, output)


@pytest.mark.skip("TODO: (#48)")
def test_summarize_removes_1_grouping(backend):
    data = data_frame(a = 1, b = 2, c = 3)
    df = backend.load_df(data)

    q1 = df >> group_by(_.a, _.b) >> summarize(n = n(_))
    assert q1.group_by == ("a")

    q2 = q1 >> summarize(n = n(_))
    assert not len(q2.group_by)


@backend_sql("TODO: pandas -  need to implement or raise this warning")
def test_summarize_no_same_call_var_refs(backend, df):
    with pytest.raises(NotImplementedError):
        df >> summarize(y = _.x.min(), z = _.y + 1)


@backend_sql
def test_summarize_removes_order_vars(backend, df):
    lazy_tbl = df >> summarize(n = n(_))

    assert not len(lazy_tbl.order_by)


@pytest.mark.skip("TODO (see #50)")
def test_summarize_unnamed_args(df):
    assert_equal_query(
            df,
            summarize(n(_)),
            pd.DataFrame({'n(_)': 4})
            )




from siuba.sql import group_by, mutate, LazyTbl, collect
from siuba.siu import _
from siuba.sql.translate import funcs


from sqlalchemy import sql
from sqlalchemy import Table, Column, Integer, String, MetaData, ForeignKey
from sqlalchemy import create_engine

from pandas.testing import assert_frame_equal
import pytest

metadata = MetaData()
users = Table('users', metadata,
    Column('id', Integer, primary_key=True),
    Column('name', String),
    Column('fullname', String),
)


addresses = Table('addresses', metadata,
    Column('id', Integer, primary_key=True),
    Column('user_id', None, ForeignKey('users.id')),
    Column('email_address', String, nullable=False)
)


@pytest.fixture(scope = "module")
def db():
    engine = create_engine('sqlite:///:memory:', echo=False)

    metadata.create_all(engine)

    conn = engine.connect()

    ins = users.insert().values(name='jack', fullname='Jack Jones')
    result = conn.execute(ins)


    ins = users.insert()
    conn.execute(ins, id=2, name='wendy', fullname='Wendy Williams')
    yield conn

# LazyTbl ---------------------------------------------------------------------

def test_lazy_tbl_table_string(db):
    tbl = LazyTbl(db, 'addresses')
    tbl.tbl.columns.user_id

def test_lazy_tbl_manual_columns(db):
    tbl = LazyTbl(db, 'addresses', columns = ('user_id', 'wrong_name'))
    tbl.tbl.columns.wrong_name
    tbl.tbl.columns.user_id

    with pytest.raises(AttributeError):
        tbl.tbl.columns.email_address

# mutate ----------------------------------------------------------------------

def test_sql_mutate(db):
    tbl = LazyTbl(db, addresses, funcs = funcs)
    f = mutate(user_id2 = _.user_id + 1)
    out1 = tbl >> f >> collect()
    out2 = tbl >> collect() >> f

    assert_frame_equal(out1, out2)


# group_by --------------------------------------------------------------------

@pytest.mark.parametrize("group_vars", [
    ["id",],                                # string syntax
    ["id", "user_id"],                      # string syntax multiple
    [_.id],                                 # _ syntax
    [_.id, _.user_id],                      # _ syntax multiple
    ])
def test_sql_group_by(db, group_vars):
    tbl = LazyTbl(db, addresses, funcs = funcs)
    group_by(tbl, *group_vars)


@pytest.mark.parametrize("group_var, error", [
    (_.id + 1, NotImplementedError),        # complex expressions
    (_.notacol, KeyError)                   # missing columns
    ])
def tets_sql_group_by_fail(db, group_var, error):
    tbl = LazyTbl(db, addresses, funcs = funcs)
    with pytest.raises(error):
        group_by(tbl, group_var)
    



# sqlvariant, allow defining 3 namespaces to override defaults
from ..translate import base_scalar, base_agg, base_nowin, SqlTranslator, win_agg
import sqlalchemy.sql.sqltypes as sa_types
from sqlalchemy import sql

scalar = SqlTranslator(
        base_scalar,
        )

aggregate = SqlTranslator(
        base_agg
        )

window = SqlTranslator(
        # TODO: should check sqlite version, since < 3.25 can't use windows
        base_nowin,
        sd = win_agg("stddev")
        )

funcs = dict(scalar = scalar, aggregate = aggregate, window = window)




"""
Note: this test file was heavily influenced by its dbplyr counterpart.

https://github.com/tidyverse/dbplyr/blob/master/tests/testthat/test-verb-mutate.R
"""
    
from siuba import _, group_by, summarize, count

import pytest
from .helpers import assert_equal_query, data_frame, backend_notimpl, backend_sql

DATA = data_frame(x = [1,2,3,4], g = ['a', 'a', 'b', 'b'])

@pytest.fixture(scope = "module")
def df(backend):
    return backend.load_df(DATA)

@pytest.mark.parametrize("query, output", [
    (count(_.g), data_frame(g = ['a', 'b'], n = [2, 2])),
    (count("g"), data_frame(g = ['a', 'b'], n = [2, 2])),
    (count("x", "g"), DATA.assign(n = 1)),
    (count(_.x, "g"), DATA.assign(n = 1))
    ])
def test_basic_count(df, query, output):
    assert_equal_query(df, query, output)


@pytest.mark.skip("TODO: sql fix unnamed expression labels in count (#69)")
def test_count_with_expression(df):
    assert_equal_query(
            df,
            count(_.x - _.x),
            pd.DataFrame({"x - x": [0], "n": [4]})
            )


@pytest.mark.skip("TODO: sql support kwargs in count (#68)")
def test_count_with_kwarg_expression(df):
    assert_equal_query(
            df,
            count(y = _.x - _.x),
            pd.DataFrame({"y": [0], "n": [4]})
            )




"""
This module holds default translations from pandas syntax to sql for 3 kinds of operations...

1. scalar - elementwise operations (e.g. array1 + array2)
2. aggregation - operations that result in a single number (e.g. array1.mean())
3. window - operations that do calculations across a window
            (e.g. array1.lag() or array1.expanding().mean())


"""

from sqlalchemy import sql
from sqlalchemy.sql import sqltypes as types
from functools import singledispatch
from .verbs import case_when, if_else
import warnings

# TODO: must make these take both tbl, col as args, since hard to find window funcs
def sa_is_window(clause):
    return isinstance(clause, sql.elements.Over) \
            or isinstance(clause, sql.elements.WithinGroup)


def sa_modify_window(clause, group_by = None, order_by = None):
    if group_by:
        group_cols = [columns[name] for name in group_by]
        partition_by = sql.elements.ClauseList(*group_cols)
        clone = clause._clone()
        clone.partition_by = partition_by

        return clone

    return clause

from sqlalchemy.sql.elements import Over
# windowed agg (group by)
# agg
# windowed scalar
# ordered set agg

class CustomOverClause: pass

class AggOver(Over, CustomOverClause):
    def set_over(self, group_by, order_by = None):
        self.partition_by = group_by
        return self


class RankOver(Over, CustomOverClause): 
    def set_over(self, group_by, order_by = None):
        self.partition_by = group_by
        return self


class CumlOver(Over, CustomOverClause):
    def set_over(self, group_by, order_by):
        self.partition_by = group_by
        self.order_by = order_by

        if not len(order_by):
            warnings.warn(
                    "No order by columns explicitly set in window function. SQL engine"
                    "does not guarantee a row ordering. Recommend using an arrange beforehand.",
                    RuntimeWarning
                    )
        return self


def win_absent(name):
    def not_implemented(*args, **kwargs):
        raise NotImplementedError("SQL dialect does not support {}.".format(name))
    
    return not_implemented

def win_over(name):
    sa_func = getattr(sql.func, name)
    return lambda col: RankOver(sa_func(), order_by = col)

def win_cumul(name):
    sa_func = getattr(sql.func, name)
    return lambda col: CumlOver(sa_func(col), rows = (None,0))

def win_agg(name):
    sa_func = getattr(sql.func, name)
    return lambda col: AggOver(sa_func(col))

def sql_agg(name):
    sa_func = getattr(sql.func, name)
    return lambda col: sa_func(col)

def sql_scalar(name, *args):
    sa_func = getattr(sql.func, name)
    return lambda col: sa_func(col, *args)

def sql_colmeth(meth, *outerargs):
    def f(col, *args):
        return getattr(col, meth)(*outerargs, *args)
    return f

def sql_astype(col, _type):
    mappings = {
            str: types.Text,
            int: types.Integer,
            float: types.Numeric,
            bool: types.Boolean
            }
    try:
        sa_type = mappings[_type]
    except KeyError:
        raise ValueError("sql astype currently only supports type objects: str, int, float, bool")
    return sql.cast(col, sa_type)

base_scalar = dict(
        # infix operators -----
        # NOTE: all sqlalchemy.Column operators are used
        # TODO: need way to implement ** operator

        # sqlalchemy.ColumnElement methods ----
        cast = sql_colmeth("cast"),
        startswith = sql_colmeth("startswith"),
        endswith = sql_colmeth("endswith"),
        between = sql_colmeth("between"),
        isin = sql_colmeth("in_"),

        # bitwise operations -----
        # TODO

        # these are methods on sql.funcs ---- 
        abs = sql_scalar("abs"),
        acos = sql_scalar("acos"),
        asin = sql_scalar("asin"),
        atan = sql_scalar("atan"),
        atan2 = sql_scalar("atan2"),
        cos = sql_scalar("cos"),
        cot = sql_scalar("cot"),
        astype = sql_astype,
        # I was lazy here and wrote lambdas directly ---
        # TODO: I think these are postgres specific?
        hour = lambda col: sql.func.date_trunc('hour', col),
        week = lambda col: sql.func.date_trunc('week', col),
        isna = sql_colmeth("is_", None),
        isnull = sql_colmeth("is_", None),
        # dply.vector funcs ----
        desc = lambda col: col.desc(),
        
        # TODO: string methods
        #str.len,
        #str.upper,
        #str.lower,
        #str.replace_all or something similar,
        #str_detect or similar,
        #str_trim func to cut text off sides
        # TODO: move to postgres specific
        n = lambda col: sql.func.count(),
        # TODO: this is to support a DictCall (e.g. used in case_when)
        dict = dict,
        # TODO: don't use singledispatch to add sql support to case_when
        case_when = case_when,
        if_else = if_else
        )

base_agg = dict(
        mean = sql_agg("avg"),
        sum = sql_agg("sum"),
        min = sql_agg("min"),
        max = sql_agg("max"),
        count = sql_agg("count"),
        # TODO: generalize case where doesn't use col
        # need better handeling of vector funcs
        # TODO: delete this, len() is not a method anywhere, or vect func
        len = lambda col: sql.func.count()
        )

base_win = dict(
        row_number = lambda col: CumlOver(sql.func.row_number()),
        min_rank = win_over("rank"),
        rank = win_over("rank"),
        dense_rank = win_over("dense_rank"),
        percent_rank = win_over("percent_rank"),
        cume_dist = win_over("cume_dist"),
        #first = win_over2("first"),
        #last = win_over2("last"),
        #nth = win_over3
        #lead = win_over4
        #lag

        # aggregate functions ---
        mean = win_agg("avg"),
        var = win_agg("variance"),
        sum = win_agg("sum"),
        min = win_agg("min"),
        max = win_agg("max"),

        # ordered set funcs ---
        #quantile
        #median

        # counts ----
        count = win_agg("count"),
        len = lambda col: sql.func.count().over(),
        #n
        #n_distinct

        # cumulative funcs ---
        #avg("id") OVER (PARTITION BY "email" ORDER BY "id" ROWS UNBOUNDED PRECEDING)
        #cummean = win_agg("
        cumsum = win_cumul("sum")
        #cummin
        #cummax

        )

# based on https://github.com/tidyverse/dbplyr/blob/master/R/backend-.R
base_nowin = dict(
        row_number   = win_absent("ROW_NUMBER"),
        min_rank     = win_absent("RANK"),
        rank         = win_absent("RANK"),
        dense_rank   = win_absent("DENSE_RANK"),
        percent_rank = win_absent("PERCENT_RANK"),
        cume_dist    = win_absent("CUME_DIST"),
        ntile        = win_absent("NTILE"),
        mean         = win_absent("AVG"),
        sd           = win_absent("SD"),
        var          = win_absent("VAR"),
        cov          = win_absent("COV"),
        cor          = win_absent("COR"),
        sum          = win_absent("SUM"),
        min          = win_absent("MIN"),
        max          = win_absent("MAX"),
        median       = win_absent("PERCENTILE_CONT"),
        quantile    = win_absent("PERCENTILE_CONT"),
        n            = win_absent("N"),
        n_distinct   = win_absent("N_DISTINCT"),
        cummean      = win_absent("MEAN"),
        cumsum       = win_absent("SUM"),
        cummin       = win_absent("MIN"),
        cummax       = win_absent("MAX"),
        nth          = win_absent("NTH_VALUE"),
        first        = win_absent("FIRST_VALUE"),
        last         = win_absent("LAST_VALUE"),
        lead         = win_absent("LEAD"),
        lag          = win_absent("LAG"),
        order_by     = win_absent("ORDER_BY"),
        str_flatten  = win_absent("STR_FLATTEN"),
        count        = win_absent("COUNT")
        )

funcs = dict(scalar = base_scalar, aggregate = base_agg, window = base_win)

# MISC ===========================================================================
# scalar, aggregate, window #no_win, agg_no_win

# local to table
# alias to LazyTbl.no_win.dense_rank...
#          LazyTbl.agg.dense_rank...

from collections.abc import Mapping
import itertools

class SqlTranslator(Mapping):
    def __init__(self, d, **kwargs):
        self.d = d
        self.kwargs = kwargs

    def __len__(self):
        return len(set(self.d) + set(self.kwargs))

    def __iter__(self):
        old_keys = iter(self.d)
        new_keys = (k for k in self.kwargs if k not in self.d)
        return itertools.chain(old_keys, new_keys)

    def __getitem__(self, x):
        try:
            return self.kwargs[x]
        except KeyError:
            return self.d[x]



from functools import singledispatch
from pandas import DataFrame
import pandas as pd
import numpy as np

from pandas.core.groupby import DataFrameGroupBy
from siuba.siu import Symbolic, Call, strip_symbolic, MetaArg, BinaryOp, create_sym_call, Lazy

DPLY_FUNCTIONS = (
        # Dply ----
        "group_by", "ungroup", 
        "select", "rename",
        "mutate", "transmute", "filter", "summarize",
        "arrange", "distinct",
        "count", "add_count",
        "head",
        # Tidy ----
        "spread", "gather",
        "nest", "unnest",
        "expand", "complete",
        # Joins ----
        "join", "inner_join", "full_join", "left_join", "right_join", "semi_join", "anti_join",
        # TODO: move to vectors
        "if_else", "case_when",
        "collect", "show_query"
        )

__all__ = [*DPLY_FUNCTIONS, "Pipeable", "pipe"]


# General TODO ================================================================
# * expressions in group_by
# * n_distinct?
# * separate_rows
# * tally

def install_pd_siu():
    # https://github.com/coursera/pandas-ply/blob/master/pandas_ply/methods.py
    func_dict = globals()
    for func_name in DPLY_FUNCTIONS:
        f = func_dict[func_name]

        method_name = "siu_{}".format(func_name)
        setattr(pd.DataFrame, method_name, f)
        setattr(DataFrameGroupBy, method_name, f)

    DataFrameGroupBy._repr_html_ = _repr_grouped_df_html_

def _repr_grouped_df_html_(self):
    return "<div><p>(grouped data frame)</p>" + self._selected_obj._repr_html_() + "</div>"

# TODO: should be a subclass of Call?
class Pipeable:
    def __init__(self, f = None, calls = None):
        if f is not None:
            if calls is not None: raise Exception()
            self.calls = [f]
        else:
            self.calls = calls

    def __rshift__(self, x):
        if isinstance(x, Pipeable):
            return Pipeable(calls = self.calls + x.calls)
        elif callable(x):
            return Pipeable(calls = self.calls + [x])

        raise Exception()

    def __rrshift__(self, x):
        if isinstance(x, Pipeable):
            return Pipeable(calls = x.calls + self.calls)
        elif callable(x):
            return Pipeable(calls = [x] + self.calls)

        return self(x)

    def __call__(self, x):
        res = x
        for f in self.calls:
            res = f(res)
        return res

pipe = Pipeable

def _regroup(df):
    # try to regroup, when user kept index (e.g. group_keys = True)
    if len(df.index.names) > 1:
        # handle cases where...
        # 1. grouping with named indices (as_index = True)
        # 2. grouping is level 0 (as_index = False)
        grp_levels = [x for x in df.index.names if x is not None] or [0]

    return df.groupby(level = grp_levels)


def simple_varname(call):
    if isinstance(call, str):
        return call

    # check for expr like _.some_var or _["some_var"]
    if (call.func in {"__getitem__", "__getattr__"}
        and isinstance(call.args[0], MetaArg)
        and isinstance(call.args[1], str)
        ):
        # return variable name
        return call.args[1]

    return None


# Symbolic Wrapper ============================================================

from functools import wraps

class NoArgs: pass

def pipe_no_args(f, cls = NoArgs):
    @f.register(cls)
    def wrapper(__data, *args, **kwargs):
        return create_pipe_call(f, MetaArg("_"), *args, **kwargs)

    return f

def register_pipe(f, cls):
    @f.register(cls)
    def wrapper(*args, **kwargs):
        return create_pipe_call(f, MetaArg("_"), *args, **kwargs)
    return f


# option: no args, custom dispatch (e.g. register NoArgs)
# strips symbols
def singledispatch2(cls, f = None):
    """Wrap singledispatch. Making sure to keep its attributes on the wrapper.
    
    This wrapper has three jobs:
        1. strip symbols off of calls
        2. pass NoArgs instance for calls like some_func(), so dispatcher can handle
        3. return a Pipeable when the first arg of a call is a symbol
    """

    # classic way of allowing args to a decorator
    if f is None:
        return lambda f: singledispatch2(cls, f)

    # initially registers func for object, so need to change to pd.DataFrame
    dispatch_func = singledispatch(f)
    if isinstance(cls, tuple):
        for c in cls: dispatch_func.register(c, f)
    else:
        dispatch_func.register(cls, f)
    # then, set the default object dispatcher to create a pipe
    register_pipe(dispatch_func, object)

    # register dispatcher for Call, and NoArgs
    pipe_call(dispatch_func)
    pipe_no_args(dispatch_func)

    @wraps(dispatch_func)
    def wrapper(*args, **kwargs):
        strip_args = map(strip_symbolic, args)
        strip_kwargs = {k: strip_symbolic(v) for k,v in kwargs.items()}

        if not args:
            return dispatch_func(NoArgs(), **strip_kwargs)

        return dispatch_func(*strip_args, **strip_kwargs)

    return wrapper


def create_pipe_call(source, *args, **kwargs):
    first, *rest = args
    return Pipeable(Call(
            "__call__",
            strip_symbolic(source),
            strip_symbolic(first),
            *(Lazy(strip_symbolic(x)) for x in rest),
            **{k: Lazy(strip_symbolic(v)) for k,v in kwargs.items()}
            ))

def pipe_with_meta(f, *args, **kwargs):
    return create_pipe_call(f, MetaArg("_"), *args, **kwargs)


def pipe_call(f):
    @f.register(Call)
    def f_dispatch(__data, *args, **kwargs):
        call = __data
        if isinstance(call, MetaArg):
            # single _ passed as first arg to function
            # e.g. mutate(_, _.id)
            return create_pipe_call(f, call, *args, **kwargs)
        else:
            # more complex _ expr passed as first arg to function
            # e.g. mutate(_.id)
            return pipe_with_meta(f, call, *args, **kwargs)

    return f


MSG_TYPE_ERROR = "The first argument to {func} must be one of: {types}"

def raise_type_error(f):
    raise TypeError(MSG_TYPE_ERROR.format(
                func = f.__name__,
                types = ", ".join(map(str, f.registry.keys()))
                ))

# Collect and show_query =========

@pipe_no_args
@singledispatch2((DataFrame, DataFrameGroupBy))
def collect(__data, *args, **kwargs):
    # simply return DataFrame, since requires no execution
    return __data


@pipe_no_args
@singledispatch2((DataFrame, DataFrameGroupBy))
def show_query(__data, simplify = False):
    print("No query to show for a DataFrame")
    return __data

# Mutate ======================================================================

# TODO: support for unnamed args
@singledispatch2(pd.DataFrame)
def mutate(__data, **kwargs):
    """Assign new variables to a DataFrame, while keeping existing ones.

    Args:
        ___data: a DataFrame
        **kwargs: new_col_name=value pairs, where value can be a function taking
                  a single argument for the data being operated on.

    Examples
    --------

    ::
        from siuba.data import mtcars
        mtcars >> mutate(cyl2 = _.cyl * 2, cyl4 = _.cyl2 * 2)
        
    """
    
    return __data.assign(**kwargs)


@mutate.register(DataFrameGroupBy)
def _mutate(__data, **kwargs):
    df = __data.apply(lambda d: d.assign(**kwargs))
    
    return _regroup(df)



# Group By ====================================================================

@singledispatch2((pd.DataFrame, DataFrameGroupBy))
def group_by(__data, *args, **kwargs):
    tmp_df = mutate(__data, **kwargs) if kwargs else __data

    by_vars = list(map(simple_varname, args))
    for ii, name in enumerate(by_vars):
        if name is None: raise Exception("group by variable %s is not a column name" %ii)

    by_vars.extend(kwargs.keys())

    return tmp_df.groupby(by = by_vars)


@singledispatch2((pd.DataFrame, DataFrameGroupBy))
def ungroup(__data):
    # TODO: can we somehow just restore the original df used to construct
    #       the groupby?
    if isinstance(__data, pd.DataFrame):
        return __data
    if isinstance(__data, pd.Series):
        return __data.reset_index()

    return __data.obj.reset_index(drop = True)



# Filter ======================================================================

@singledispatch2(pd.DataFrame)
def filter(__data, *args):
    """Keep rows where conditions are true.

    Args:
        ___data: a DataFrame
        *args: conditions that must be met to keep a column. Multiple conditions
               are combined using ``&``.

    Examples
    --------

    ::
        from siuba.data import mtcars
        # keep rows where cyl is 4 and mpg is less than 25
        mtcars >> filter(mtcars, _.cyl ==  4, _.mpg < 25) 

    """
    crnt_indx = True
    for arg in args:
        crnt_indx &= arg(__data) if callable(arg) else arg

    # use loc or iloc to subset, depending on crnt_indx ----
    # the main issue here is that loc can't remove all rows using a slice
    # and iloc can't use a boolean series
    if isinstance(crnt_indx, bool) or isinstance(crnt_indx, np.bool_):
        # iloc can do slice, but not a bool series
        return __data.iloc[slice(None) if crnt_indx else slice(0),:]

    return __data.loc[crnt_indx,:]

@filter.register(DataFrameGroupBy)
def _filter(__data, *args):
    df_filter = filter.registry[pd.DataFrame]

    filtered = __data.apply(df_filter, *args)

    return _regroup(filtered)



# Summarize ===================================================================

@singledispatch2(DataFrame)
def summarize(__data, **kwargs):
    """Assign variables that are single number summaries of a DataFrame.


    Args:
        ___data: a DataFrame
        **kwargs: new_col_name=value pairs, where value can be a function taking
                  a single argument for the data being operated on.

    Note
    ----

    Grouped DataFrames will produce one row for each group. Ungrouped DataFrames
    will produce a single row.


    Examples
    --------

    ::
        from siuba.data import mtcars
        mtcars >> summarize(mean = _.disp.mean(), n = n(_))
        
    """
    results = {}
    for k, v in kwargs.items():
        res = v(__data) if callable(v) else v

        # TODO: validation?

        results[k] = res
        
    return DataFrame(results, index = [0])

    
@summarize.register(DataFrameGroupBy)
def _summarize(__data, **kwargs):
    df_summarize = summarize.registry[pd.DataFrame]

    df = __data.apply(df_summarize, **kwargs)
        
    group_by_lvls = list(range(df.index.nlevels - 1))
    out = df.reset_index(group_by_lvls)
    out.index = pd.RangeIndex(df.shape[0])

    return out



# Transmute ===================================================================

@singledispatch2(DataFrame)
def transmute(__data, *args, **kwargs):
    arg_vars = list(map(simple_varname, args))
    for ii, name in enumerate(arg_vars):
        if name is None: raise Exception("complex, unnamed expression at pos %s not supported"%ii)

    f_mutate = mutate.registry[pd.DataFrame]

    df = f_mutate(__data, **kwargs) 

    return df[[*arg_vars, *kwargs.keys()]]

@transmute.register(DataFrameGroupBy)
def _transmute(__data, *args, **kwargs):
    # Note: not the most efficient function. Selects columns every group.
    f_transmute = transmute.registry[pd.DataFrame]

    df = __data.apply(f_transmute, *args, **kwargs)

    return _regroup(df)



# Select ======================================================================

from collections import OrderedDict
from itertools import chain

class Var:
    def __init__(self, name, negated = False, alias = None):
        self.name = name
        self.negated = negated
        self.alias = alias

    def __neg__(self):
        return self.to_copy(negated = not self.negated)

    def __eq__(self, x):
        name = x.name if isinstance(x, Var) else x
        return self.to_copy(name = name, negated = False, alias = self.name)

    def __call__(self, *args, **kwargs):
        call = Call('__call__',
                    BinaryOp('__getattr__', MetaArg("_"), self.name),
                    *args,
                    **kwargs
                    )

        return self.to_copy(name = call)


    def __repr__(self):
        return "Var('{self.name}', negated = {self.negated}, alias = {self.alias})" \
                    .format(self = self)

    def __str__(self):
        op = "-" if self.negated else ""
        pref = self.alias + " = " if self.alias else ""
        return "{pref}{op}{self.name}".format(pref = pref, op = op, self = self)

    def to_copy(self, **kwargs):
        return self.__class__(**{**self.__dict__, **kwargs})


class VarList:
    def __getattr__(self, x):
        return Var(x)

    def __getitem__(self, x):
        return Var(x)


def var_slice(colnames, x):
    """Return indices in colnames correspnding to start and stop of slice."""
    # TODO: produces bahavior similar to df.loc[:, "V1":"V3"], but can reverse
    # TODO: make DRY
    # TODO: reverse not including end points
    if isinstance(x.start, Var):
        start_indx = (colnames == x.start.name).idxmax()
    elif isinstance(x.start, str):
        start_indx = (colnames == x.start).idxmax()
    else:
        start_indx = x.start or 0

    if isinstance(x.stop, Var):
        stop_indx = (colnames == x.stop.name).idxmax() + 1
    elif isinstance(x.stop, str):
        stop_indx = (colnames == x.stop).idxmax() + 1
    else:
        stop_indx = x.stop or len(colnames)

    if start_indx > stop_indx:
        return stop_indx, start_indx
    else:
        return start_indx, stop_indx

def var_put_cols(name, var, cols):
    if isinstance(name, list) and var.alias is not None:
        raise Exception("Cannot assign name to multiple columns")
    
    names = [name] if not isinstance(name, list) else name

    for name in names:
        if var.negated:
            if name in cols: cols.pop(name)
        #elif name in cols: cols.move_to_end(name)
        else: cols[name] = var.alias

def flatten_var(var):
    if isinstance(var, Var) and isinstance(var.name, (tuple, list)):
        return [var.to_copy(name = x) for x in var.name]
    
    return [var]
            



def var_select(colnames, *args):
    # TODO: don't erase named column if included again
    colnames = colnames if isinstance(colnames, pd.Series) else pd.Series(colnames)
    cols = OrderedDict()
    everything = None

    #flat_args = var_flatten(args)
    all_vars = chain(*map(flatten_var, args))

    # Add entries in pandas.rename style {"orig_name": "new_name"}
    for arg in all_vars:
        # strings are added directly
        if isinstance(arg, str):
            cols[arg] = None
        # integers add colname at corresponding index
        elif isinstance(arg, int):
            cols[colnames[arg]] = None
        # general var handling
        elif isinstance(arg, Var):
            # remove negated Vars, otherwise include them
            if arg.negated and everything is None:
                # first time using negation, apply an implicit everything
                everything = True
                cols.update((k, None) for k in colnames if k not in cols)

            # slicing can refer to single, or range of columns
            if isinstance(arg.name, slice):
                start, stop = var_slice(colnames, arg.name)
                for ii in range(start, stop):
                    var_put_cols(colnames[ii], arg, cols)
            # method calls like endswith()
            elif callable(arg.name):
                # TODO: not sure if this is a good idea...
                #       basically proxies to pandas str methods (they must return bool array)
                indx = arg.name(colnames.str)
                var_put_cols(colnames[indx].tolist(), arg, cols)
                #cols.update((x, None) for x in set(colnames[indx]) - set(cols))
            else:
                var_put_cols(arg.name, arg, cols)
        else:
            raise Exception("variable must be either a string or Var instance")

    return cols

def var_create(*args):
    vl = VarList()
    all_vars = []
    for arg in args:
        if callable(arg) and not isinstance(arg, Var):
            all_vars.append(arg(vl))
        else:
            all_vars.append(arg)
     
    return all_vars

@singledispatch2(DataFrame)
def select(__data, *args, **kwargs):
    if kwargs:
        raise NotImplementedError(
                "Using kwargs in select not currently supported. "
                "Use _.newname == _.oldname instead"
                )
    var_list = var_create(*args)

    od = var_select(__data.columns, *var_list)

    to_rename = {k: v for k,v in od.items() if v is not None}

    return __data[list(od)].rename(columns = to_rename)
    
@select.register(DataFrameGroupBy)
def _select(__data, *args, **kwargs):
    raise Exception("Selecting columns of grouped DataFrame currently not allowed")



# Rename ======================================================================

@singledispatch2(DataFrame)
def rename(__data, **kwargs):
    # TODO: allow names with spaces, etc..
    col_names = {simple_varname(v):k for k,v in kwargs.items()}
    if None in col_names:
        raise ValueError("Rename needs column name (e.g. 'a' or _.a), but received %s"%col_names[None])

    return __data.rename(columns  = col_names)

@rename.register(DataFrameGroupBy)
def _rename(__data, **kwargs):
    raise NotImplementedError("Selecting columns of grouped DataFrame currently not allowed")



# Arrange =====================================================================

def _call_strip_ascending(f):
    if isinstance(f, Symbolic):
        f = f.source

    if isinstance(f, Call) and f.func == "__neg__":
        return f.args[0], False

    return f, True

@singledispatch2(DataFrame)
def arrange(__data, *args):
    # TODO:
    #   - add arguments to pass to sort_values (e.g. ascending, kind)
    # 
    # basically need some (1) select behavior, (2) mutate-like behavior
    # df.sort_values is the obvious candidate, but only takes names, not expressions
    # to work around this, we make a shallow copy of data, and add sorting columns
    # then drop them at the end
    # 
    # sort order is determined by using a unary w/ Call e.g. -_.repo

    df = __data.copy(deep = False)
    n_cols = len(df.columns)
    n_args = len(args)

    #kwargs = {n_cols + ii: arg for ii,arg in enumerate(args)}

    # TODO: more careful handling of arg types (true across library :/ )..
    tmp_cols = []
    sort_cols = []
    ascending = []
    for ii, arg in enumerate(args):
        f, asc = _call_strip_ascending(arg)
        ascending.append(asc)

        col = simple_varname(f)
        if col is not None:
            sort_cols.append(col)
        else:
            # TODO: could screw up if user has columns names that are ints...
            sort_cols.append(n_cols + ii)
            tmp_cols.append(n_cols + ii)

            df[n_cols + ii] = f(df)


    return df.sort_values(by = sort_cols, kind = "mergesort", ascending = ascending) \
             .drop(tmp_cols, axis = 1)


@arrange.register(DataFrameGroupBy)
def _arrange(__data, *args):
    raise NotImplementedError("TODO: arrange with grouped DataFrame")



# Distinct ====================================================================

@singledispatch2(DataFrame)
def distinct(__data, *args, _keep_all = False, **kwargs):
    # using dict as ordered set
    cols = {simple_varname(x): True for x in args}
    if None in cols:
        raise Exception("positional arguments must be simple column, "
                        "e.g. _.colname or _['colname']"
                        )

    # mutate kwargs
    cols.update(kwargs)

    # special case: use all variables when none are specified
    if not len(cols): cols = __data.columns

    tmp_data = mutate(__data, **kwargs).drop_duplicates(list(cols)).reset_index(drop = True)

    if not _keep_all:
        return tmp_data[list(cols)]

    return tmp_data
        
@distinct.register(DataFrameGroupBy)
def _distinct(__data, *args, _keep_all = False, **kwargs):
    df = __data.apply(lambda x: distinct(x, *args, _keep_all = _keep_all, **kwargs))
    return _regroup(df)


# if_else
# TODO: move to vector.py
@singledispatch
def if_else(__data, *args, **kwargs):
    raise_type_error(__data)

@if_else.register(Call)
@if_else.register(Symbolic)
def _if_else(__data, *args, **kwargs):
    return create_sym_call(if_else, __data, *args, **kwargs)

@if_else.register(pd.Series)
def _if_else(cond, true_vals, false_vals):
    true_indx = np.where(cond)[0]
    false_indx = np.where(~cond)[0]

    result = np.repeat(None, len(cond))

    result[true_indx] =  true_vals[true_indx] if np.ndim(true_vals) else true_vals
    result[false_indx] = false_vals[false_indx] if np.ndim(false_vals) else false_vals

    # TODO: inefficient way to downcast?
    return np.array(list(result))


# case_when ----------------
# note that here, we don't use @Pipeable.add_to_dispatcher.
# because case_when takes a dictionary of cases, we need to wrap cases into
# a Call, so that it can be handled by call tree visitors, etc..
# TODO: evaluate this non-table verb approach
from siuba.siu import DictCall

def _val_call(call, data, n, indx = None):
    if not callable(call):
        return call

    arr = call(data)
    if arr.shape != (n,):
        raise ValueError("Expected call to return array of shape {}"
                         "but it returned shape {}".format(n, arr.shape))

    return arr[indx] if indx is not None else arr


@singledispatch2((pd.DataFrame,pd.Series))
def case_when(__data, cases):
    if isinstance(cases, Call):
        cases = cases(__data)
    # TODO: handle when receive list of (k,v) pairs for py < 3.5 compat?

    stripped_cases = {strip_symbolic(k): strip_symbolic(v) for k,v in cases.items()}
    n = len(__data)
    out = np.repeat(None, n)
    for k, v in reversed(list(stripped_cases.items())):
        if callable(k):
            result = _val_call(k, __data, n)
            indx = np.where(result)[0]

            val_res = _val_call(v, __data, n, indx)
            out[indx] = val_res
        elif k:
            # e.g. k is just True, etc..
            val_res = _val_call(v, __data, n)
            out[:] = val_res

    # by recreating an array, attempts to cast as best dtype
    return np.array(list(out))

@case_when.register(Symbolic)
@case_when.register(Call)
def _case_when(__data, cases):
    if not isinstance(cases, dict):
        raise Exception("Cases must be a dictionary")
    dict_entries = dict((strip_symbolic(k), strip_symbolic(v)) for k,v in cases.items())
    cases_arg = Lazy(DictCall("__call__", dict, dict_entries))
    return create_sym_call(case_when, __data, cases_arg)




# Count =======================================================================

def _count_group(data, *args):
    crnt_cols = set(data.columns)
    out_col = "n"
    while out_col in crnt_cols: out_col = out_col + "n"

    return 


@singledispatch2(pd.DataFrame)
def count(__data, *args, wt = None, sort = False, **kwargs):
    # TODO: if expr, works like mutate

    #group by args
    if wt is None:
        counts = group_by(__data, *args, **kwargs).size().reset_index()
    else:
        wt_col = simple_varname(wt)
        if wt_col is None:
            raise Exception("wt argument has to be simple column name")
        counts = group_by(__data, *args, **kwargs)[wt_col].sum().reset_index()


    # count col named, n. If that col already exists, add more "n"s...
    crnt_cols = set(counts.columns)
    out_col = "n"
    while out_col in crnt_cols: out_col = out_col + "n"

    # rename the tally column to correct name
    counts.rename(columns = {counts.columns[-1]: out_col}, inplace = True)

    if sort:
        return counts.sort_values(out_col, ascending = False)

    return counts


@singledispatch2(pd.DataFrame)
def add_count(__data, *args, wt = None, sort = False, **kwargs):
    counts = count(__data, *args, wt = wt, sort = sort, **kwargs)

    on = list(counts.columns)[:-1]
    return __data.merge(counts, on = on)
    


# Tally =======================================================================

# Nest ========================================================================

@singledispatch2(pd.DataFrame)
def nest(__data, *args, key = "data"):
    """Nest columns within a DataFrame.
    

    Args:
        ___data: a DataFrame
        *args: the names of columns to be nested. May use any syntax used by
               the ``select`` function.
        key: the name of the column that will hold the nested columns.

    Examples
    --------

    ::
        from siuba.data import mtcars
        mtcars >> nest(-_.cyl)
        
    """
    # TODO: copied from select function
    var_list = var_create(*args)
    od = var_select(__data.columns, *var_list)

    # unselected columns are treated similar to using groupby
    grp_keys = list(k for k in __data.columns if k not in set(od))
    nest_keys = list(od)

    # AFAICT you can't name the col created in the apply here
    # but it might be more efficient to act on the groupby obj directly
    out = __data.groupby(grp_keys).apply(lambda d: [d[nest_keys]]).reset_index()
    out.rename(columns = {out.columns[-1]: key}, inplace = True)
    # hack, can assign list of dataframes, but not in a groupby?
    # it ended up making a series of lists each with a single dataframe
    out[key] = [x[0] for x in out[key]]

    return out

@nest.register(DataFrameGroupBy)
def _nest(__data, *args, key = "data"):
    grp_keys = [x.name for x in __data.grouper.groupings]
    if None in grp_keys:
        raise NotImplementedError("All groupby variables must be named when using nest")

    return nest(__data.obj, -Var(grp_keys), *args, key = key)




# Unnest ======================================================================

@singledispatch2(pd.DataFrame)
def unnest(__data, key = "data"):
    """Unnest a column holding nested data (e.g. Series of lists or DataFrames).
    
    Args:
        ___data: a DataFrame
        key: the name of the column to be unnested.

    Examples
    --------

    ::
        import pandas as pd
        df = pd.DataFrame({'id': [1,2], 'data': [['a', 'b'], ['c', 'd']]})
        df >> unnest()
        
    """
    # TODO: currently only takes key, not expressions
    nrows_nested = __data[key].apply(len, convert_dtype = True)
    indx_nested = nrows_nested.index.repeat(nrows_nested)

    grp_keys = list(__data.columns[__data.columns != key])

    # flatten nested data
    data_entries = map(_convert_nested_entry, __data[key])
    long_data = pd.concat(data_entries, ignore_index = True)
    long_data.name = key

    # may be a better approach using a multi-index
    long_grp = __data.loc[indx_nested, grp_keys].reset_index(drop = True)
    
    return long_grp.join(long_data)

def _convert_nested_entry(x):
    if isinstance(x, (tuple, list)):
        return pd.Series(x)
    
    return x


# Joins =======================================================================
from collections.abc import Mapping
from functools import partial

# TODO: will need to use multiple dispatch
@singledispatch2(pd.DataFrame)
def join(left, right, on = None, how = None):
    if not isinstance(right, DataFrame):
        raise Exception("right hand table must be a DataFrame")
    if how is None:
        raise Exception("Must specify how argument")

    if isinstance(on, Mapping):
        left_on, right_on = zip(*on.items())
        return left.merge(right, how = how, left_on = left_on, right_on = right_on)

    return left.merge(right, how = how, on = on)


@join.register(object)
def _join(left, right, on = None, how = None):
    raise Exception("Unsupported type %s" %type(left))


@singledispatch2(pd.DataFrame)
def semi_join(left, right = None, on = None):
    if isinstance(on, Mapping):
        left_on, right_on = zip(*on.items())
        return left.merge(right[right_on], how = 'inner', left_on = left_on, right_on = right_on)

    if on is None:
        on_cols = set(left.columns).intersection(set(right.columns))
        if not len(on_cols):
            raise Exception("No joining column specified, and no shared column names")
    elif isinstance(on, str):
        on_cols = [on]
    else:
        on_cols = on

    return left.merge(right.loc[:,on_cols], how = 'inner', on = on_cols)

@singledispatch2(pd.DataFrame)
def anti_join(left, right = None, on = None):
    raise NotImplementedError("anti_join not currently implemented")

left_join = partial(join, how = "left")
right_join = partial(join, how = "right")
full_join = partial(join, how = "full")
inner_join = partial(join, how = "inner")


# Head ========================================================================

@singledispatch2(pd.DataFrame)
def head(__data, n = 5):
    return __data.head(n)

# Gather ======================================================================

@singledispatch2(pd.DataFrame)
def gather(__data, key = "key", value = "value", *args, drop_na = False, convert = False):
    # TODO: implement var selection over *args
    if convert:
        raise NotImplementedError("convert not yet implemented")

    # TODO: copied from nest and select
    var_list = var_create(*args)
    od = var_select(__data.columns, *var_list)

    value_vars = list(od) or None

    id_vars = [col for col in __data.columns if col not in od]
    long = pd.melt(__data, id_vars, value_vars, key, value)

    if drop_na:
        return long[~long[value].isna()]

    return long



# Spread ======================================================================

@singledispatch2(pd.DataFrame)
def spread(__data, key, value, fill = None, reset_index = True):
    id_cols = [col for col in __data.columns if col not in {key, value}]
    wide = __data.set_index(id_cols + [key]).unstack(level = -1)
    
    if fill is not None:
        wide.fillna(fill, inplace = True)
    
    # remove multi-index from both rows and cols
    wide.columns = wide.columns.droplevel().rename(None)
    if reset_index:
        wide.reset_index(inplace = True)
    
    return wide

# Expand/Complete ====================================================================
from pandas.core.reshape.util import cartesian_product

@singledispatch2(pd.DataFrame)
def expand(__data, *args, fill = None):
    var_names = list(map(simple_varname, args))
    cols = [__data[name] for name in var_names]
    # see https://stackoverflow.com/a/25636395/1144523
    cprod = cartesian_product(cols)
    expanded = pd.DataFrame(np.array(cprod).T)
    expanded.columns = var_names

    return expanded


@singledispatch2(pd.DataFrame)
def complete(__data, *args, fill = None):
    expanded = expand(__data, *args, fill = fill)

    # TODO: should we attempt to coerce cols back to original types?
    #       e.g. NAs will turn int -> float
    on_cols = list(expanded.columns)
    df = __data.merge(expanded, how = "right", on = on_cols)
    
    if fill is not None:
        for col_name, val in fill.items():
            df[col_name].fillna(val, inplace = True)

    return df
    

# Install Siu =================================================================

install_pd_siu()



# sqlvariant, allow defining 3 namespaces to override defaults
from ..translate import (
        base_scalar, base_agg, base_win, SqlTranslator, 
        win_agg, sql_scalar
        )
import sqlalchemy.sql.sqltypes as sa_types
from sqlalchemy import sql

def sql_log(col, base = None):
    if base is None:
        return sql.func.ln(col)
    return sql.func.log(col)

def sql_round(col, n):
    return sql.func.round(sql.cast(col, sa_types.Numeric()), n)

def sql_str_contains(col, pat, case, *args, **kwargs):
    if args or kwargs:
        raise NotImplementedError("Only pat and case arg of contains allowed.")

    infix = "~" if case else "~*"

    return col.op(infix, pat)

# handle when others is a list?
def sql_str_cat(col, others, sep, join = None):
    if join is not None:
        raise NotImplementedError("join argument of cat not supported")


scalar = SqlTranslator(
        base_scalar,
        log = sql_log,
        round = sql_round,
        contains = sql_str_contains,
        year = lambda col: sql.func.extract('year', sql.cast(col, sql.sqltypes.Date)),
        concat = sql.func.concat,
        cat = sql.func.concat,
        str_c = sql.func.concat
        )

aggregate = SqlTranslator(
        base_agg
        )

window = SqlTranslator(
        base_win,
        any = win_agg("bool_or"),
        all = win_agg("bool_and"),
        lag = win_agg("lag")
        )

funcs = dict(scalar = scalar, aggregate = aggregate, window = window)



import pytest
from siuba.dply.verbs import mutate
from siuba.siu import _

import pandas as pd
from pandas.testing import assert_frame_equal

@pytest.fixture(scope = "function")
def df1():
    yield pd.DataFrame({
        "repo": ["pandas", "dplyr", "ggplot2", "plotnine"],
        "owner": ["pandas-dev", "tidyverse", "tidyverse", "has2k1"],
        "language": ["python", "R", "R", "python"],
        "stars": [17800, 2800, 3500, 1450],
        "x": [1,2,3,None]
        })

def test_dply_mutate(df1):
    op_stars_1k = lambda d: d.stars * 1000
    out1 = mutate(df1, stars_1k = op_stars_1k)
    out2 = df1.assign(stars_1k = op_stars_1k)

    assert_frame_equal(out1, out2)

def test_dply_mutate_sym(df1):
    op_stars_1k = _.stars * 1000
    out1 = mutate(df1, stars_1k = op_stars_1k)
    out2 = df1.assign(stars_1k = op_stars_1k)

    assert_frame_equal(out1, out2)

# VarList and friends ------

from siuba.dply.verbs import flatten_var, Var, VarList

def test_flatten_vars():
    v_a, v_b = flatten_var(-Var(["a", "b"]))
    assert v_a.name == "a"
    assert v_b.name == "b"
    assert all([v_a.negated, v_b.negated])

@pytest.mark.parametrize("x", [
    ["x"],
    [Var("x")],
    [slice(0,1)]
    ])
def test_flatten_vars_noop(x):
    assert x is flatten_var(x)[0]

def test_VarList_getitem_siu():
    vl = VarList()
    f = lambda _: _[_.a, _.b]
    var = f(vl)
    v_a, v_b = var.name
    assert v_a.name == "a"
    assert v_b.name == "b"

def test_VarList_getitem():
    vl = VarList()
    var = vl["a":"b", "c"]
    assert isinstance(var.name[0], slice)
    assert var.name[1] == "c"



# Select ----------------------------------------------------------------------

from siuba.dply.verbs import select

def test_varlist_multi_slice(df1):
    out = select(df1, lambda _: _["repo", "owner"])
    assert out.columns.tolist() == ["repo", "owner"]

def test_varlist_multi_slice_negate(df1):
    out = select(df1, lambda _: -_["repo", "owner"])
    assert out.columns.tolist() == ["language", "stars", "x"]


# Distinct --------------------------------------------------------------------

from siuba.dply.verbs import distinct

def test_distinct_no_args():
    df =pd.DataFrame({'x': [1,1,2], 'y': [1,1,2]})
    assert_frame_equal(distinct(df), df.drop_duplicates().reset_index(drop = True))


# Nest ------------------------------------------------------------------------

from siuba.dply.verbs import nest, unnest

def test_nest_grouped(df1):
    out = nest(df1.groupby("language"))
    assert out.columns[0] == "language"

    entry = out.loc[0, "data"]
    assert isinstance(entry, pd.DataFrame)
    assert entry.shape == (2, df1.shape[1] - 1)
    
def test_nest_unnest(df1):
    # important that repo is first column, since grouping vars moved to front
    out = unnest(nest(df1, -Var("repo")))

    # since a nest implies grouping, we need to sort before comparing
    sorted_df = df1.sort_values(["repo"]).reset_index(drop = True)
    assert_frame_equal(out, sorted_df)

def test_unnest_lists():
    df = pd.DataFrame({'id': [1,2], 'data': [['a'], ['x', 'y']]})
    out = unnest(df)
    assert_frame_equal(
            out,
            pd.DataFrame({'id': [1,2,2], 'data': ['a','x','y']})
            )





import pandas as pd
import numpy as np
from ..siu import create_sym_call, Symbolic
from functools import singledispatch

# TODO: move into siu
def register_symbolic(f):
    @f.register(Symbolic)
    def _dispatch_symbol(__data, *args, **kwargs):
        return create_sym_call(f, __data.source, *args, **kwargs)

    return f


# fct_reorder -----------------------------------------------------------------

@register_symbolic
@singledispatch
def fct_reorder(fct, x, func = np.median):
    x_vals = x.values if isinstance(x, pd.Series) else x
    s = pd.Series(x_vals, index = fct)

    # for each cat, calc agg func, make values of ordered the codes
    ordered = s.groupby(level = 0).agg(func).sort_values()
    ordered[:] = np.arange(len(ordered))
    codes = ordered[s.index.values]
    return pd.Categorical.from_codes(codes, list(ordered.index))


# fct_recode ------------------------------------------------------------------

@register_symbolic
@singledispatch
def fct_recode(fct, **kwargs):
    if not isinstance(fct, pd.Categorical):
        fct = pd.Categorical(fct)

    rev_kwargs = {v:k for k,v in kwargs.items()}
    return fct.rename_categories(rev_kwargs)


# fct_collapse ----------------------------------------------------------------

@register_symbolic
@singledispatch
def fct_collapse(fct, recat, group_other = None):
    if not isinstance(fct, pd.Categorical):
        fct = pd.Categorical(fct)

    # each existing cat will map to a new one ----
    # need to know existing to new cat
    # need to know new cat to new code
    cat_to_new = {k: None for k in fct.categories}
    new_cat_set = {k: True for k in fct.categories} 
    for new_name, v in recat.items():
        v = [v] if not np.ndim(v) else v
        for old_name in v:
            if cat_to_new[old_name] is not None:
                raise Exception("category %s was already re-assigned"%old_name)
            cat_to_new[old_name] = new_name
            del new_cat_set[old_name]
            new_cat_set[new_name] = True    # add new cat

    # collapse all unspecified cats to group_other if specified ----
    for k, v in cat_to_new.items():
        if v is None:
            if group_other is not None:
                new_cat_set[group_other] = True
                cat_to_new[k] = group_other
                del new_cat_set[k]
            else:
                cat_to_new[k] = k

    # map from old cat to new code ----
    # calculate new codes
    new_cat_set = {k: ii for ii, k in enumerate(new_cat_set)}
    # map old cats to them
    remap_code = {old: new_cat_set[new] for old, new in cat_to_new.items()}

    new_codes = fct.map(remap_code)
    new_cats = list(new_cat_set.keys())
    return pd.Categorical.from_codes(new_codes, new_cats)


# fct_lump --------------------------------------------------------------------

@register_symbolic
@singledispatch
def fct_lump(fct, n = None, prop = None, w = None, other_level = "Other", ties = None):
    if ties is not None:
        raise NotImplementedError("ties is not implemented")
    
    if n is None and prop is None:
        raise NotImplementedError("Either n or prop must be specified")

    if prop is not None:
        raise NotImplementedError("prop arg is not implemented")

    keep_cats = _fct_lump_n_cats(fct, n, w, other_level, ties)
    return fct_collapse(fct, {k:k for k in keep_cats}, group_other = other_level) 

def _fct_lump_n_cats(fct, n, w, other_level, ties):
    # TODO: currently always selects n, even if ties
    ascending = n < 0
    arr = _get_values(w) if w is not None else 1
    ser = pd.Series(arr, index = fct)
    sorted_arr = ser.groupby(level = 0).sum().sort_values(ascending = ascending)
    return sorted_arr.iloc[:abs(n)].index.values

def _get_values(x):
    # TODO: move into utility, note pandas now encouraging .array method
    if isinstance(x, pd.Series): return x.values

    return x


# fct_rev ---------------------------------------------------------------------

@register_symbolic
@singledispatch
def fct_rev(fct):
    if not isinstance(fct, pd.Categorical):
        fct = pd.Categorical(fct)

    rev_levels = list(reversed(fct.categories))

    return fct.reorder_categories(rev_levels)
    
